{
  "type": "record",
  "name": "ExecutionRequestResult",
  "namespace": "com.linkedin.pegasus2avro.execution",
  "doc": "The result of an execution request",
  "fields": [
    {
      "name": "status",
      "type": "string",
      "doc": "The status of the execution request"
    },
    {
      "name": "report",
      "type": [
        "null",
        "string"
      ],
      "doc": "The pretty-printed execution report.",
      "default": null
    },
    {
      "name": "startTimeMs",
      "type": [
        "null",
        "long"
      ],
      "doc": "Time at which the request was created",
      "default": null,
      "Searchable": {
        "fieldName": "startTimeMs",
        "fieldType": "COUNT",
        "queryByDefault": false
      }
    },
    {
      "name": "durationMs",
      "type": [
        "null",
        "long"
      ],
      "doc": "Duration in milliseconds",
      "default": null
    }
  ],
  "Aspect": {
    "name": "dataHubExecutionRequestResult"
  }
}