{
  "type": "record",
  "name": "DataHubSecretValue",
  "namespace": "com.linkedin.pegasus2avro.secret",
  "doc": "The value of a DataHub Secret",
  "fields": [
    {
      "name": "name",
      "type": "string",
      "doc": "The display name for the secret",
      "Searchable": {
        "fieldType": "TEXT_PARTIAL"
      }
    },
    {
      "name": "value",
      "type": "string",
      "doc": "The AES-encrypted value of the DataHub secret."
    },
    {
      "name": "description",
      "type": [
        "null",
        "string"
      ],
      "doc": "Description of the secret",
      "default": null
    }
  ],
  "Aspect": {
    "name": "dataHubSecretValue"
  }
}